################################################################################
#
# Copyright 2006-2007 Novell, Inc.
# Author: Bart Wise <bwise@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
#
# Class and associated subclasses to deal with graphics adapters
# for various hypervisors.
#
# Class Inheritance Tree:
#   GraphicsAdapter
#       XenGraphicsAdapter
#       LibVirtGraphicsAdapter
#           KVMGraphicsAdapter
#           XenGraphicsAdapter (future?)
################################################################################


################################################################################
# GLOBAL DECLARAIONS
################################################################################
graphics_adapter    = None


################################################################################
# CLASS: GraphicsAdapter
#
# Class to abstract the graphics adapters for various hypervisors
# and configurations.
################################################################################
class GraphicsAdapter:

    _ParaVirtualGraphicsAdapterList  = []
    _FullyVirtualGraphicsAdapterList = []
    _GraphicsAdapterList             = []


    #---------------------------------------------------------------------------
    # Class:  GraphicsAdapter
    # Method: __init__
    # Parms:
    #   None.
    #
    # Initializes the class.
    #---------------------------------------------------------------------------
    def __init__(self):
        self._GraphicsAdapterList = self._FullyVirtualGraphicsAdapterList + self._ParaVirtualGraphicsAdapterList
        #
        # Verify that 'none' is part of the list.
        #
        #TODO: put none in the list.


    #---------------------------------------------------------------------------
    # Class:  GraphicsAdapter
    # Method: get_fv_graphics_adapter_list
    # Parms:
    #   None.
    #
    # Returns the fully-virtual graphics adapter list.
    #---------------------------------------------------------------------------
    def get_fv_graphics_adapter_list(self):
        return _FullyVirtualGraphicsAdapterList


    #---------------------------------------------------------------------------
    # Class:  GraphicsAdapter
    # Method: get_graphics_adapter_list
    # Parms:
    #   None.
    #
    # Returns the complete graphics adapter list.
    #---------------------------------------------------------------------------
    def get_graphics_adapter_list(self):
        return _GraphicsAdapterList


    #---------------------------------------------------------------------------
    # Class:  GraphicsAdapter
    # Method: get_pv_graphics_adapter_list
    # Parms:
    #   None.
    #
    # Returns the paravirtual graphics adapter list.
    #---------------------------------------------------------------------------
    def get_pv_graphics_adapter_list(self):
        return _ParaVirtualGraphicsAdapterList


################################################################################
# CLASS: XenGraphicsAdapter
#
# Class to abstract the console.
################################################################################
class XenGraphicsAdapter(GraphicsAdapter):
    #---------------------------------------------------------------------------
    # Class:  KVMGraphicsAdapter
    # Method: __init__
    # Parms:
    #   None.
    #
    # Initializes the class.
    #---------------------------------------------------------------------------
    def __init__(self):
        self._FullyVirtualGraphicsAdapterList = ['cirrus', 'vesa']
        self._ParaVirtualGraphicsAdapterList = ['para']

        #
        # Aways call the parent __init__ function after we 
        # have initialized for this subclass.
        #
        GraphicsAdapter.__init__(self)


################################################################################
# CLASS: LibVirtGraphicsAdapter
#
# Class to abstract the console.
################################################################################
class LibVirtGraphicsAdapter(GraphicsAdapter):
    #---------------------------------------------------------------------------
    # Class:  LibVirtGraphicsAdapter
    # Method: __init__
    # Parms:
    #   None.
    #
    # Initializes the class.
    #---------------------------------------------------------------------------
    def __init__(self):
        #
        # Aways call the parent __init__ function after we 
        # have initialized for this subclass.
        #
        GraphicsAdapter.__init__(self)


################################################################################
# CLASS: KVMGraphicsAdapter
#
# Class to abstract the console.
################################################################################
class KVMGraphicsAdapter(LibVirtGraphicsAdapter):
    #---------------------------------------------------------------------------
    # Class:  KVMGraphicsAdapter
    # Method: __init__
    # Parms:
    #   None.
    #
    # Initializes the class.
    #---------------------------------------------------------------------------
    def __init__(self):
        self._FullyVirtualGraphicsAdapterList = ['cirrus', 'std', 'vmware']

        #
        # Aways call the parent __init__ function after we 
        # have initialized for this subclass.
        #
        LibVirtGraphicsAdapter.__init__(self)




