# GObject wrapper of installation job
#
# Copyright 2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

from gi.repository import GObject
import threading
import vminstall.log

class GJob(GObject.GObject):
    """This class starts a new thread for the VM creation job.  Events
    during the creation are emitted as GTK signals.
    """

    __gsignals__ = {
        'install-progress-change': (GObject.SignalFlags.RUN_FIRST, None, (str, float)),
        'install-progress-close': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'install-show-vm': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        'install-show-error': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'install-done': (GObject.SignalFlags.RUN_FIRST, None, ())
    }

    def __init__(self, job, defaults, errors):
        GObject.GObject.__init__(self)
        self.job = job
        self.defaults = defaults
        self.errors = errors

    def start(self):
        def _getSettingsCB(job):
            return self.defaults
        def _errorCB(job, text):
            self.errors.insert(0, text)
            self.emit('install-show-error')
        def _stateChangeCB(job, state):
            self.emit('install-progress-change', state[2], -1.0)
        def _progressChangeCB(job, fraction):
            self.emit('install-progress-change', '', fraction)
        def _consoleCB(job):
            self.emit('install-show-vm', job.guest.uuid)
        def _finish_worker():
            try:
                try:
                    self.job.run(GetSettingsCB=_getSettingsCB,
                                 StateChangeCB=_stateChangeCB,
                                 ConsoleCB=_consoleCB,
                                 ErrorCB=_errorCB,
                                 ProgressCB=_progressChangeCB)
                except Exception as e:
                    vminstall.log.error("VM creation job failed: %s" % str(e))
                    self.emit('install-progress-close')
                    self.errors.insert(0, e)
                    self.emit('install-show-error')
                else:
                    vminstall.log.info("VM creation job finished.")
                    self.emit('install-progress-close')
            finally:
                self.emit('install-done')
        thread = threading.Thread(target=_finish_worker)
        thread.setDaemon(True)
        thread.start()

