# Sound Interface Card
#
# Copyright 2011 Novell, Inc.
# Author: 
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

from .exceptions import *
from . import msg
from . import caps

################################################################################
# GLOBAL DATA
################################################################################

#
# Define a dictionary with the valid sound cards
#
xen_sound_cards = {
    'none': msg.none,
    'ac97': msg.sound_hw_ac97,
    'es1370': msg.sound_hw_es1370,
    'ich6': msg.sound_hw_ich6,
    'sb16': msg.sound_hw_sb16,
}

kvm_sound_cards = {
    'none': msg.none,
    'ac97': msg.sound_hw_ac97,
    'es1370': msg.sound_hw_es1370,
    'ich6': msg.sound_hw_ich6,
    'sb16': msg.sound_hw_sb16,
}

################################################################################
# GLOBAL FUNCTIONS
################################################################################

#-------------------------------------------------------------------------------
# Function: parse_string
# Parms:
#   card:    type of sound card
#
# The input string is parsed and validated
#-------------------------------------------------------------------------------
def parse_string(card):
    if card not in sound_cards:
        raise ParamError(err=ParamError.E_INVALID, details=card)
    return card

#-------------------------------------------------------------------------------
# Function: get_sound_cards
# Parms:
#
# The input string is parsed and validated
#-------------------------------------------------------------------------------
def get_sound_cards():
    if caps.is_xen():
        return xen_sound_cards
    else:
        return kvm_sound_cards

def get_sound(current_card=None):
    if current_card is None:
        current_card = 'none'
    return get_sound_cards()[current_card]
